"""test for convergence checks"""

import numpy as np
import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))
import pyqlaw


def test_check_convergence():
    oe = np.array([1.0, 0.1, 0.2, 0.3, 0.4])
    oeT = oe + 1e-6 * np.ones(5,)
    woe = np.ones(5,)
    tol_loose = 1e-3 * np.ones(5,)
    tol_tight = 1e-7 * np.ones(5,)

    assert pyqlaw.check_convergence_keplerian(oe, oeT, woe, tol_loose) == True
    assert pyqlaw.check_convergence_keplerian(oe, oeT, woe, tol_tight) == False

    assert pyqlaw.check_convergence_keplerian.py_func(oe, oeT, woe, tol_loose) == True
    assert pyqlaw.check_convergence_keplerian.py_func(oe, oeT, woe, tol_tight) == False

    assert pyqlaw.check_convergence_mee(oe, oeT, woe, tol_loose) == True
    assert pyqlaw.check_convergence_mee(oe, oeT, woe, tol_tight) == False

    assert pyqlaw.check_convergence_mee.py_func(oe, oeT, woe, tol_loose) == True
    assert pyqlaw.check_convergence_mee.py_func(oe, oeT, woe, tol_tight) == False

    oe_unsafe = np.array([1.0, 0.0, 0.2, 0.3, 0.4, 0.5])
    oe_min = np.array([0.1, 1e-3, 0.0, 0.0, 0.0, 0.0])
    oe_max = np.array([9.0, 0.9, np.pi, 2*np.pi, 2*np.pi, 2*np.pi])
    oe_clean = pyqlaw.elements_safety.py_func(oe_unsafe, oe_min, oe_max)
    return

if __name__ == "__main__":
    test_check_convergence()
    print("Done!")


"""Tests for elements helper functions"""


import numpy as np
from numpy.random import rand
import matplotlib.pyplot as plt

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))
import pyqlaw


def diff_kep(KEP0, KEP1):
    return np.array([
        KEP1[0] - KEP0[0],  # a
        KEP1[1] - KEP0[1],  # e
        KEP1[2] - KEP0[2],  # i
        np.arccos(np.cos(KEP1[3] - KEP0[3])),  # RAAN
        np.arccos(np.cos(KEP1[4] - KEP0[4])),  # omega
        np.arccos(np.cos(KEP1[5] - KEP0[5])),  # ta
    ])


def test_kep_sv():
    # initial and final elements: [a,e,i,RAAN,omega,ta]
    LU = 42164.0
    GM_EARTH = 398600.44
    VU = np.sqrt(GM_EARTH/LU)
    TU = LU/VU

    rp_gto = 200 + 6378
    ra_gto = 35786 + 6378
    sma_gto = (rp_gto + ra_gto)/(2*LU)
    ecc_gto = (ra_gto - rp_gto)/(ra_gto + rp_gto)

    KEP0s = [
        np.array([sma_gto,1e-3,np.deg2rad(23),0.3,np.deg2rad(100),3.06]),
        np.array([sma_gto,ecc_gto,np.deg2rad(23),0.3,np.deg2rad(100),3.06]),
        np.array([sma_gto,ecc_gto,np.deg2rad(23),0.3,np.deg2rad(200),3.06]),
    ]
    
    for KEP0 in KEP0s:
        # convert to SV and back
        sv = pyqlaw.kep2sv(KEP0, GM_EARTH)
        kep_converted_back = pyqlaw.sv2kep(sv, GM_EARTH)

        # jit-compiled ver
        assert np.abs(pyqlaw.get_semiMajorAxis(sv, GM_EARTH) - KEP0[0]) <= 1e-12
        assert np.abs(np.linalg.norm(pyqlaw.get_eccentricity(sv, GM_EARTH)) - KEP0[1]) <= 1e-12
        assert np.abs(pyqlaw.get_inclination(sv)             - KEP0[2]) <= 1e-12
        assert np.abs(pyqlaw.get_raan(sv)                    - KEP0[3]) <= 1e-12
        assert np.abs(pyqlaw.get_omega(sv, GM_EARTH)         - KEP0[4]) <= 1e-12
        assert np.abs(pyqlaw.get_trueanom(sv, GM_EARTH)      - KEP0[5]) <= 1e-12
        assert np.abs(pyqlaw.get_period(sv, GM_EARTH) - 2*np.pi*np.sqrt(KEP0[0]**3/GM_EARTH)) <= 1e-12
        
        # pure python ver for coverage purposes
        assert np.abs(pyqlaw.get_semiMajorAxis.py_func(sv, GM_EARTH) - KEP0[0]) <= 1e-12
        assert np.abs(np.linalg.norm(pyqlaw.get_eccentricity.py_func(sv, GM_EARTH)) - KEP0[1]) <= 1e-12
        assert np.abs(pyqlaw.get_inclination.py_func(sv)             - KEP0[2]) <= 1e-12
        assert np.abs(pyqlaw.get_raan.py_func(sv)                    - KEP0[3]) <= 1e-12
        assert np.abs(pyqlaw.get_omega.py_func(sv, GM_EARTH)         - KEP0[4]) <= 1e-12
        assert np.abs(pyqlaw.get_trueanom.py_func(sv, GM_EARTH)      - KEP0[5]) <= 1e-12
        assert np.abs(pyqlaw.get_period.py_func(sv, GM_EARTH) - 2*np.pi*np.sqrt(KEP0[0]**3/GM_EARTH)) <= 1e-12
        
        # jit-compiled ver
        sv = pyqlaw.kep2sv(KEP0, GM_EARTH)
        kep_converted_back = pyqlaw.sv2kep(sv, GM_EARTH)
        assert all(np.abs(diff_kep(kep_converted_back, KEP0)) <= 1e-12)
        
        # pure python ver for coverage purposes
        sv = pyqlaw.kep2sv.py_func(KEP0, GM_EARTH)
        kep_converted_back = pyqlaw.sv2kep.py_func(sv, GM_EARTH)
        assert all(np.abs(diff_kep(kep_converted_back, KEP0)) <= 1e-12)

        # conversion to MEE
        MEE0 = pyqlaw.kep2mee(KEP0)
        MEE0_nojit = pyqlaw.kep2mee.py_func(KEP0)
        assert all(np.abs(MEE0 - MEE0_nojit) <= 1e-15)

        KEP0_back = pyqlaw.mee2kep(MEE0)
        assert all(np.abs(diff_kep(KEP0_back, KEP0)) <= 1e-12)

        KEP0_back_nojit = pyqlaw.mee2kep.py_func(MEE0)
        assert all(np.abs(diff_kep(KEP0_back_nojit, KEP0)) <= 1e-12)

        # anomaly conversions
        EA0_direct_nojit = pyqlaw.ta2ea.py_func(KEP0[5], KEP0[1])
        EA0_nojit = pyqlaw.mee2ea.py_func(MEE0)
        assert np.abs(EA0_direct_nojit - EA0_nojit) <= 1e-15
    return
    

def test_kep_mee_a():
    # initial and final elements: [a,e,i,RAAN,omega,ta]
    LU = 42164.0
    GM_EARTH = 398600.44
    VU = np.sqrt(GM_EARTH/LU)
    TU = LU/VU
    MU = 1.0

    rp_gto = 200 + 6378
    ra_gto = 35786 + 6378
    sma_gto = (rp_gto + ra_gto)/(2*LU)
    ecc_gto = (ra_gto - rp_gto)/(ra_gto + rp_gto)
    KEP0 = np.array([sma_gto,ecc_gto,np.deg2rad(23),0.3,2.1,0.0])
    MEEA0 = pyqlaw.kep2mee_with_a(KEP0)
    KEP_BACK = pyqlaw.mee_with_a2kep(MEEA0)
    assert all(np.abs(diff_kep(KEP_BACK, KEP0)) <= 1e-12)

    MEEA0_thru_MEE = pyqlaw.mee2mee_with_a.py_func(pyqlaw.kep2mee(KEP0))
    assert all(np.abs(MEEA0_thru_MEE - MEEA0) <= 1e-12)

    # jit-compiled ver
    kep_converted_back = pyqlaw.mee_with_a2kep(MEEA0)
    assert all(np.abs(diff_kep(kep_converted_back, KEP0)) <= 1e-12)

    # pure python ver for coverage purposes
    kep_converted_back = pyqlaw.mee_with_a2kep.py_func(pyqlaw.kep2mee_with_a.py_func(KEP0))
    assert all(np.abs(diff_kep(kep_converted_back, KEP0)) <= 1e-12)

    # get orbital states in Cartesian state over one rev
    coord_jit = pyqlaw.get_orbit_coordinates(KEP0,MU,15)
    coord     = pyqlaw.get_orbit_coordinates.py_func(KEP0,MU,15)
    assert all(np.abs(coord[:,0] - coord_jit[:,0]) <= 1e-12)

    RV0 = pyqlaw.mee_with_a2sv(MEEA0, MU)
    assert all(np.abs(coord[:,0] - RV0) <= 1e-15)

    # pure python ver for coverage purposes
    RV0_nojit = pyqlaw.mee_with_a2sv.py_func(MEEA0, MU)
    assert all(np.abs(coord[:,0] - RV0_nojit) <= 1e-15)


if __name__ == "__main__":
    test_kep_sv()
    test_kep_mee_a()
    print("Done!")


"""Test for runge-kutta algos"""

import numpy as np
from numpy.random import rand
import matplotlib.pyplot as plt

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))
import pyqlaw


def test_rk4_rk45():
    # construct problem
    prob = pyqlaw.QLaw(
        elements_type="keplerian",
        integrator="rk4"
    )#, verbosity=2)
    # initial and final elements: [a,e,i,RAAN,omega,ta]
    oe0 = np.array([1.5, 0.2, 0.3, 1e-2, 1e-2, 1e-2])
    oeT = np.array([2.2, 0.3, 1.1, 0.3, 0.0])
    woe = [1.0, 1.0, 1.0, 1.0, 1.0]
    # spacecraft parameters
    mass0 = 1.0
    tmax = 1e-3
    mdot = 1e-4
    tf_max = 3000.0
    t_step = 1.0
    # set problem
    prob.set_problem(oe0, oeT, mass0, tmax, mdot, tf_max, t_step, woe=woe)
    
    # query psi
    accel_thrust = tmax/mass0
    _, _, _, psi, _, _ = pyqlaw.lyapunov_control_angles(
        fun_lyapunov_control=prob.lyap_fun,
        mu=prob.mu, 
        f=accel_thrust, 
        oe=oe0, 
        oeT=prob.oeT, 
        rpmin=prob.rpmin, 
        m_petro=prob.m_petro, 
        n_petro=prob.n_petro, 
        r_petro=prob.r_petro, 
        b_petro=prob.b_petro, 
        k_petro=prob.k_petro, 
        wp=prob.wp, 
        woe=prob.woe,
    )
    u = np.array([0.1,0.4,0.2])
    u *= accel_thrust/np.linalg.norm(u)
    ode_params = (prob.mu, u, psi[0], psi[1], psi[2], np.zeros(3,))

    # known check case
    doe_check = np.array([
        0.00392952589873677, 0.00209742303058534, 
        0.0004363521324628918, 2.9535057990649755e-05, 
        -0.001241443160063603, 0.8345326718595107])
    
    h = 0.01    # step-size

    # evaluate eom (jit-ed)
    ynext_rk4 = pyqlaw.rk4(pyqlaw.eom_kep_gauss, 0.0, h, oe0, ode_params)
    ynext_rk45, h_next = pyqlaw.rkf45(pyqlaw.eom_kep_gauss, 0.0, h, oe0, ode_params)
    assert all(np.abs(ynext_rk4[0:4] - ynext_rk45[0:4]) <= 1e-6)
    assert np.abs(ynext_rk4[5] - ynext_rk45[5]) <= 1e-5

    # evaluate eom (pure python)
    ynext_rk4 = pyqlaw.rk4.py_func(pyqlaw.eom_kep_gauss, 0.0, h, oe0, ode_params)
    ynext_rk45, h_next = pyqlaw.rkf45.py_func(pyqlaw.eom_kep_gauss, 0.0, h, oe0, ode_params)
    assert all(np.abs(ynext_rk4[0:4] - ynext_rk45[0:4]) <= 1e-6)
    assert np.abs(ynext_rk4[5] - ynext_rk45[5]) <= 1e-5
    return 


if __name__=="__main__":
    test_rk4_rk45()
    print("Done!")


"""
Check convergence
"""

import copy
import numpy as np
from numba import njit

@njit
def check_convergence_keplerian(oe, oeT, woe, tol_oe):
    """Check convergence between oe and oeT for Keplerian elements"""
    check_array = np.zeros(len(tol_oe),)
    doe = np.abs(np.array([
        oeT[0] - oe[0],  # a
        oeT[1] - oe[1],  # e
        oeT[2] - oe[2],  # i
        np.arccos(np.cos(oeT[3] - oe[3])),  # RAAN
        np.arccos(np.cos(oeT[4] - oe[4])),  # omega
    ]))
    for idx in range(len(tol_oe)):
        if woe[idx] == 0.0:
            check_array[idx] = 1        # consider converged
        else:
            check_array[idx] = doe[idx] <= tol_oe[idx]

    if np.sum(check_array)==len(tol_oe):
        return True
    else:
        return False


@njit
def check_convergence_mee(oe, oeT, woe, tol_oe):
    """Check convergence between oe and oeT for MEE or MEE-a"""
    check_array = np.zeros(len(tol_oe),)
    doe = np.abs(oeT[0:5] - oe[0:5])
    for idx in range(len(tol_oe)):
        if woe[idx] == 0.0:
            check_array[idx] = 1        # consider converged
        else:
            check_array[idx] = doe[idx] <= tol_oe[idx]

    if np.sum(check_array)==len(tol_oe):
        return True
    else:
        return False


@njit
def elements_safety(oe, oe_min, oe_max):
    """Ensure osculating elements stays within bounds
    
    Args:
        oe (np.array): current osculating elements
        oe_min (np.array): minimum values for each osculating element
        oe_max (np.array): maximum values for each osculating element

    Returns:
        (np.array): "cleaned" osculating elements
    """
    oe_clean = oe
    for idx in range(5):
        oe_clean[idx] = min(max(oe_min[idx], oe[idx]), oe_max[idx])
    return oe_clean


"""Object containing perturbation computation scheme"""

from numba import njit
import numpy as np
import spiceypy as spice

from ._elements import kep2sv, mee_with_a2sv


@njit
def perturbation_third_body_battin(r, s, mu_third):
    """Third body perturbation acceleration via Battin's formulation
    
    Args:
        r (np.array): position vector of spacecraft w.r.t. primary body
        s (np.array): position vector of third body w.r.t. primary body
        mu_third (float): GM of third body

    Returns:
        (np.array): third-body perturbation acceleration
    """
    d = r - s
    dnorm = np.sqrt(np.dot(d,d))
    q = np.dot(r, r - 2*s)/np.dot(s,s)
    F = q*((3 + 3*q + q**2)/(1 + np.sqrt(1+q)**3))
    return -mu_third/dnorm**3 * (r + F*s)


@njit
def perturbation_J2(rECI, T_ECI2PA, mu, J2, Re):
    """Compute J2 perturbation in perifocal frame and return vector in ECI frame
    
    Args:
        rECI (np.array): position vector in ECI frame
        T_ECI2PA (np.array): transformation matrix from ECI to principal axes frame
        mu (float): GM of central body
        J2 (float): J2 coefficient of central body
        Re (float): equatorial radius of central body

    Returns:
        (np.array): J2 perturbation acceleration in ECI frame
    """
    r = np.linalg.norm(rECI)
    coeff = -3*mu*J2*Re**2/(2*r**5)
    rPA = T_ECI2PA @ rECI
    a_J2_PA = coeff * np.array([
        (1 - 5*rPA[2]**2) * rPA[0],
        (1 - 5*rPA[2]**2) * rPA[1],
        (3 - 5*rPA[2]**2) * rPA[2]
    ])
    return T_ECI2PA.T @ a_J2_PA


def perturbation_SRP(r_sun2sc, AU, k_srp):
    """Compute solar radiation pressure (SRP) perturbation
    
    Args:
        r_sun2sc (np.array): position vector of spacecraft w.r.t. Sun, in LU
        AU (float): Astronomical unit, in LU
        k_srp (float): acceleration magnitude, LU/TU^2

    Returns:
        (np.array): SRP acceleration
    """
    return k_srp * AU**2  * r_sun2sc / np.linalg.norm(r_sun2sc)**3


@njit
def pxformECI2RTN(rECI, vECI):
    """Compute 3-by-3 transformation matrix from ECI to RTN frame
    
    Args:
        rECI (np.array): position vector in ECI frame
        vECI (np.array): velocity vector in ECI frame

    Returns:
        (np.array): 3-by-3 transformation matrix from ECI to RTN frame
    """
    er = rECI / np.linalg.norm(rECI)
    h = np.cross(rECI, vECI)
    en = h / np.linalg.norm(h)
    et = np.cross(en, er)
    
    ECI2RTN = np.zeros((3,3))
    ECI2RTN[0,:] = er
    ECI2RTN[1,:] = et
    ECI2RTN[2,:] = en
    return ECI2RTN


class SpicePerturbations:
    """Class contains parameters for computing perturbations in RTN frame.
    SPICE is used to query positions of third bodies and transformation matrices.
    Default inputs are for an Earth-centered orbit.
    
    Args:
        et_ref (float): reference time in ET, i.e. ephemeris time corresponding to `t = 0`
        LU (float): length unit in km
        TU (float): time unit in seconds
        obs_id (str): SPICE ID of observer body, i.e. central body
        frame_qlaw (str): SPICE frame in which Q-law is defined
        frame_PA (str): SPICE frame corresponding to principal axes frame of central body
        third_bodies_names (list): list of SPICE IDs of third bodies
        third_bodies_gms (list): list of GM's of third bodies
        J2 (float): J2 coefficient of central body
        Re_km (float): equatorial radius of central body in km
        use_J2 (bool): flag to use J2 perturbation
    """
    def __init__(
        self,
        et_ref,
        LU,
        TU,
        obs_id = "399",
        frame_qlaw = "J2000",
        frame_PA = "ITRF93",
        third_bodies_names = ["301", "10"], 
        third_bodies_gms = None,
        J2 = 0.00108263,
        Re_km = 6378.0,
        P_SRP_SI = 4.56e-6,
        Cr_SRP = 1.2,
        Am_SI = 0.01,
        use_J2 = True,
        use_SRP = True,
    ):
        # store parameters
        self.et_ref = et_ref
        self.LU = LU
        self.TU = TU
        self.obs_id = obs_id
        self.frame_qlaw = frame_qlaw
        self.frame_PA = frame_PA
        self.third_bodies_names = third_bodies_names
        self.J2 = J2
        self.Re = Re_km / LU
        self.k_srp = P_SRP_SI * Cr_SRP * Am_SI / (1e3 * LU/TU**2)

        # flags for turning perturbation terms on/off
        self.use_J2 = use_J2
        self.use_SRP = use_SRP

        # get GM's if not provided
        VU = LU / TU
        MU_REF = LU * VU**2
        self.obs_mu = spice.bodvrd(obs_id, "GM", 1)[1][0] / MU_REF
        if (third_bodies_gms is None) and (third_bodies_names is not None):
            self.third_bodies_gms = [spice.bodvrd(ID, "GM", 1)[1][0]/MU_REF for ID in third_bodies_names]
        else:
            self.third_bodies_gms = third_bodies_gms
        return
    
    def get_perturbations_RTN(self, t, oe, elements_type):
        """Compute perturbation in RTN frame
        
        Args:
            t (float): time in TU
            oe (np.array): osculating elements
            elements_type (str): type of osculating elements

        Returns:
            (np.array): perturbation acceleration in RTN frame
        """
        # get inertial position & velocity
        if elements_type == "keplerian":
            rv_ECI = kep2sv(oe, self.obs_mu)
        elif elements_type == "mee_with_a":
            rv_ECI = mee_with_a2sv(oe, self.obs_mu)
        else:
            raise ValueError(f"Invalid elements type {elements_type}")

        # first compute perturbations in ECI
        acc_ptrb_ECI = np.zeros(3,)

        # third-body perturbation in ECI
        et = self.et_ref + t * self.TU     # time to query spice ephemeris
        for mu3,ID in zip(self.third_bodies_gms, self.third_bodies_names):
            pos3_km,_ = spice.spkpos(ID, et, self.frame_qlaw, "NONE", self.obs_id)
            acc_ptrb_ECI += perturbation_third_body_battin(rv_ECI[0:3], pos3_km/self.LU, mu3)

        # J2 perturbation in ECI
        if self.use_J2:
            T_ECI2PA = spice.pxform(self.frame_qlaw, self.frame_PA, et)
            acc_ptrb_ECI += perturbation_J2(rv_ECI[0:3], T_ECI2PA, self.obs_mu, self.J2, self.Re)

        # SRP perturbation in ECI
        if self.use_SRP:
            pos_sun_km,_ = spice.spkpos("10", et, self.frame_qlaw, "NONE", self.obs_id)
            r_sun2sc = rv_ECI[0:3] - pos_sun_km/self.LU
            acc_ptrb_ECI += perturbation_SRP(r_sun2sc, spice.convrt(1.0, "AU", "km")/self.LU, self.k_srp)

        # transform perturbations to RTN
        T_ECI2RTN = pxformECI2RTN(rv_ECI[0:3], rv_ECI[3:6])
        acc_ptrb_RTN = T_ECI2RTN @ acc_ptrb_ECI
        return acc_ptrb_RTN